// SWAGGER - TEST HELPERS
// =============================================================================
// This file includes definitions and methods which can be used to rapidly test
// Swagger API middlewares and handlers.

import { expect, tv4 } from "chai";

import { utility as TKUtility } from "@troveng/turbokit";

// JavaScript doesn't understand such fancy terminology, so we have to map
// the Swagger spec's expected types to things that JS does support.
export const swaggerTypes =
  { integer  : "number"
  , long     : "number"
  , float    : "number"
  , double   : "number"
  , string   : "string"
  , byte     : "string"
  , binary   : "string"
  , boolean  : "boolean"
  , date     : "string"
  , dateTime : "string"
  , password : "string"
  };

function validateParameter ( message, param, value ) {
  it( message, () => {
    expect( value ).to.exist;

    if ( param.type ) {
      // Because the value passed to validateParameter may have been run through
      // a function like decodeURIComponent, it is likely to be cast as a string
      // even if the spec calls for it to be a different type. Assuming the spec
      // defines the parameter's type, we can coerce it into something that
      // allows the appropriate comparison.

      // We also want to do this first
      switch ( swaggerTypes[ param.type ] ) {
        case "number":
          value = Number( value );
          break;

        case "boolean":
          value = Boolean( value );
          break;
      }

      expect( value ).to.be.a( swaggerTypes[ param.type ] );
    }

    if ( param.enum ) {
      expect( param.enum ).to.contain( value );
    }

    if ( param.schema ) {
      // We know that everything this validator will be expected to test
      // is going to be sending JSON, so JSON.parse should always be safe.

      // Even though tv4 is being provided by a chai plugin, this is a
      // bad place to use it: the default behavior is to give a standard
      // error message that indicates the JSON was invalid, but not
      // where or why
      const validation = tv4.validateResult( JSON.parse( value ), param.schema );

      if ( validation.error && validation.error.message ) {
        throw new Error( validation.error.message );
      }
    }
  });
}

export function validateAgainstSpec ( path, query, init, spec, options = {} ) {
  options =
    { "Content-Type" : false
    , Authorization  : false
    , header         : false
    , query          : false
    , path           : false
    , body           : false
    , ...options
    };

  query = TKUtility.paramStringToObject( query );

  it( "has valid Headers", () => {
    expect( init ).to.have.property( "headers" )
      .that.is.an.instanceof( Headers );
  });

  if ( options["Content-Type"] ) {
    it( "uses an appropriate Content-Type", () => {
      expect( spec.consumes ).to.include( init.headers.get( "Content-Type" ) );
    });
  }

  if ( options.Authorization ) {
    it( "contains an Authorization token", () => {
      const authToken = init.headers.get( "Authorization" ).split( /\s/ );

      expect( authToken ).to.have.length( 2 );
      expect( authToken[0] ).to.equal( "bearer" );
      expect( authToken[1] ).to.be.a( "string" );
    });
  }

  spec.parameters.forEach( param => {
    // Each of these test cases is mainly triggered by three different criteria:
    // 1. Is the option to test it flagged true?
    // 2. Is the parameter required?
    // 3. Is the parameter not required, but present anyways?
    switch ( param.in ) {
      case "header":
        if ( options.header &&
           ( param.required || Boolean( init.headers.get( param.name ) ) )
           ) {
          validateParameter( `has a header with a defined ${ param.name }`
                           , param
                           , init.headers.get( param.name )
                           );
        }
        return;

      case "path":
        // TODO: Wait to implement this until we have a definite use case
        return;

      case "query":
        if ( options.query &&
           ( param.required || Boolean( query[ param.name ] ) )
           ) {
          validateParameter( `has valid query parameter for ${ param.name }`
                           , param
                           , query[ param.name ]
                           );
        }
        return;

      case "body":
        if ( options.body &&
           ( param.required || Boolean( init.body ) )
           ) {
          const schemaName = param.schema && param.schema.$ref ||
                             param.schema.items && param.schema.items.$ref;

          validateParameter( `has a valid body (${ schemaName })`
                           , param
                           , init.body
                           );
        }
        return;
    }
  });
}
