// WEBSERVER
// =============================================================================

// Removing this flag causes certain conditional logic to be tripped, preparing
// the app for server-side rendering. In example, the Redux store will not try
// to collect dehydrated state from an object on `window`, and React components
// will not attempt to require() images.
delete process.env.BROWSER;

if (process.env.FORMATTED_ERRORS === "true") {
  // Optionally improve the legibility of node errors. This must be disabled in live environments
  // due to how it conflicts with Sumologic.
  require( "pretty-error" ).start();
}

// POLYFILL FETCH
require( "isomorphic-fetch" );

const path = require( "path" );
const compression = require( "compression" );
const express = require( "express" );
const webpack = require( "webpack" );
const favicon = require( "serve-favicon" );
const Cookies = require( "cookies" );
const config = require("./config");
const app = express();

// In the case of NODE_ENV=staging or undefined default to production.
let nodeEnv = process.env.NODE_ENV;
nodeEnv = nodeEnv === "staging" || !nodeEnv  ? "production" : nodeEnv;

process.on("unhandledRejection", rejection => console.error( rejection ) );
console.info(`Starting webserver with NODE_ENV=${nodeEnv}` );

// HELPER METHODS
// =====================
function ensureSecure(req, res, next) {
  // ELB healthcheck requests do not include the header `x-forwarded-proto` and
  // they do not need to be redirected. Thus this ignores any requests with the
  // user-agent of the health checker.
  if (req.get("user-agent").startsWith("ELB-HealthChecker")) {
    return res.sendStatus(200);
  }

  // ELB sets this header based on the protocol of the originally forwarded request.
  if (req.get("x-forwarded-proto") === "https"){
    return next();
  }

  res.redirect("https://" + req.hostname + req.url);
}

// EXPRESS CONFIGURATION
// =====================
app.use(Cookies.express());

if ( process.env.DISABLE_HTTPS !== "true" && nodeEnv === "production" ) {
  console.log('Enabling HTTPS redirect');
  app.all('*', ensureSecure); // must be at the top of all routing calls
}

if (process.env.DISABLE_HTTPS === "true") {
  console.log('Disabling HTTPS redirect');
}

app.use( compression() );
app.use( favicon( path.join( __dirname, "..", "public", "favicons", "favicon.ico" ) ) );
app.use( "/public/", express.static( path.join( __dirname, "..", "public" ) ) );

// LIVE DEVELOPMENT MODE
if ( nodeEnv === "development" ) {
  const config   = require( "../webpack/webpack.config.development" );
  const compiler = webpack( config );

  app.use( require( "webpack-dev-middleware" )( compiler
    , { noInfo             : true
      , stats              : "errors-only"
      , publicPath         : config[0].output.publicPath
      , historyApiFallback : true
    }
  ));

  app.use( require( "webpack-hot-middleware" )(
    compiler.compilers.find( compiler => compiler.name === "client" ), {
      log: () => {}
    })
  );
  app.use( require( "webpack-hot-server-middleware" )( compiler ) );
} else if ( nodeEnv === "production" ) {
  const stats  = require( "../dist/stats.generated.json" );
  const serverRenderer = require( "../dist/server" ).default;

  app.use( "/assets/", express.static( path.join( __dirname, "..", "dist", "assets" ) ) );
  app.use( serverRenderer( stats ) );
  app.use( ( err, req, res ) => {
    console.error( err.stack );
    res.status(500).send( "Internal server error: 500" );
  });
} else {
  return new Error( `A mode of "${ nodeEnv }" is not supported` );
}

app.listen( config.port, function ( error ) {
  if ( error ) {
    return console.error( error );
  }

  if ( nodeEnv === "development" ) {
    console.info( "Webpack build running, please wait..." );
  }
});